from dateutil.parser import parse
from datetime import datetime, timedelta
from pathlib import Path
import numpy as np
import asyncio
import logging
import sys
import os
from aiohttp_requests import requests
from typing import Sequence, Union, Iterator, Dict

TIMEOUT = 600  # arbitrary, seconds

if sys.version_info < (3, 7):
    raise RuntimeError('Python >= 3.7 required')


async def urlretrieve(url: str, fn: Path, overwrite: bool = False):
    if not overwrite and fn.is_file() and fn.stat().st_size > 10000:
        print(f'SKIPPED {fn}')
        return
# %% download
    R = await requests.get(url, allow_redirects=True, timeout=TIMEOUT)

    if R.status != 200:
        logging.error(f'could not download {url}  {R.status}')
        return

    print(url)

    data = await R.read()

    with fn.open('wb') as f:
        f.write(data)


def download(treq: Sequence[Union[str, datetime]],
             site: Sequence[str], odir: Path,
             urls: Dict[str, str], overwrite: bool = False):
    """
    concurrent download video and calibration files

    Parameters
    ----------

    treq : datetime or list of datatime
        time request e.g. '2012-02-01T03' or a time range ['2012-02-01T03', '2012-02-01T05']
    site : str or list of str
        camera site e.g. gako, fykn
    odir : pathlib.Path
        output directory e.g. ~/data
    overwrite : bool
        overwrite existing files--normally wasteful of download space, unless you had a corrupted download.
    urls : dict of str
        stem of URLs to [web]site(s) hosting files
    """

# %% sanity checks
    odir = Path(odir).expanduser().resolve()
    odir.mkdir(parents=True, exist_ok=True)

    if isinstance(site, str):
        site = [site]

    if isinstance(treq, (str, datetime)):
        treq = [treq]

    start = parse(treq[0]) if isinstance(treq[0], str) else treq[0]
    if len(treq) == 2:
        end = parse(treq[1]) if isinstance(treq[1], str) else treq[1]
    else:
        end = start

    if end < start:
        raise ValueError('start time must be before end time!')
# %% start download
    if os.name == 'nt' and sys.version_info < (3, 8):
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())  # type: ignore

    asyncio.run(arbiter(site, start, end, odir, overwrite, urls))


async def arbiter(sites: Sequence[str],
                  start: datetime, end: datetime,
                  odir: Path, overwrite: bool,
                  urls: Dict[str, str]):
    """
    This starts len(sites) tasks concurrently.
    Thus if you only have one site, it only downloads one file at a time.

    A more elegant way is to setup a task pool.
    However normally we are downloading N sites across the same timespan,
    where N is typically in the 3..10 range or so.

    Parameters
    ----------
    sites : str or list of str
        sites to download video from
    start : datetime
        starting time
    end : datetime
        ending time
    odir : Path
        where to download data to
    overwrite : bool, optional
        overwrite existing data
    urls : dict of str
        sites hosting data


    Example of task pool:
    https://github.com/ec500-software-engineering/asyncio-subprocess-ffmpeg/blob/c82d3a243078d8e865740cd9c3328fe7fd6ea52e/asyncioffmpeg/ffplay.py#L45

    """

    futures = [_download_cal(site, odir, urls['cal_stem'], overwrite) for site in sites]

    await asyncio.gather(*futures)

    futures = [_download_video(site, odir, start, end, urls['video_stem'], overwrite) for site in sites]

    await asyncio.gather(*futures)


async def _download_video(site: str, odir: Path,
                          start: datetime, end: datetime,
                          url_stem: str, overwrite: bool):

    for url in _urlgen(site, start, end, url_stem):
        print('GEN: ', url)
        await urlretrieve(url, odir / url.split('/')[-1], overwrite)


def _urlgen(site: str, start: datetime, end: datetime, url_stem: str) -> Iterator[str]:

    for T in np.arange(start, end+timedelta(hours=1), timedelta(hours=1)):
        t = T.astype(datetime)
        fpath = (f'{url_stem}{site}/{t.year:4d}/{t.month:02d}/'
                 f'thg_l1_asf_{site}_{t.year:4d}{t.month:02d}{t.day:02d}{t.hour:02d}_v01.cdf')

        yield fpath


async def _download_cal(site: str, odir: Path, url_stem: str,
                        overwrite: bool = False):

    fpath = f"{url_stem}thg_l2_asc_{site}_19700101_v01.cdf"

    await urlretrieve(fpath,  odir / fpath.split('/')[-1], overwrite)



#!/usr/bin/env python
import themisasi.web as tw
import pytest
from pathlib import Path
import requests.exceptions

R = Path(__file__).parent

VIDEO_BASE = 'http://themis.ssl.berkeley.edu/data/themis/thg/l1/asi/'
CAL_BASE = 'http://themis.ssl.berkeley.edu/data/themis/thg/l2/asi/cal/'
urls = {'video_stem': VIDEO_BASE, 'cal_stem': CAL_BASE}


def test_single_time_site():
    try:
        tw.download('2006-09-29T14', 'gako', R, urls)
    except requests.exceptions.ConnectionError:
        pytest.xfail('bad internet connection')

    assert (R/'thg_l1_asf_gako_2006092914_v01.cdf').is_file()
    assert (R/'thg_l2_asc_gako_19700101_v01.cdf').is_file()


@pytest.mark.asyncio
async def test_cal():
    try:
        await tw._download_cal('inuv', R, urls['cal_stem'])
    except requests.exceptions.ConnectionError:
        pytest.xfail('bad internet connection')


def test_time_range():
    try:
        tw.download(('2006-09-29T14', '2006-09-30-04'), 'gako', R, urls)
    except requests.exceptions.ConnectionError:
        pytest.xfail('bad internet connection')

    assert (R/'thg_l1_asf_gako_2006093004_v01.cdf').is_file()
    assert (R/'thg_l2_asc_gako_19700101_v01.cdf').is_file()


def test_multi_site():
    try:
        tw.download('2006-09-30-04', ['gako', 'fykn'], R, urls)
    except requests.exceptions.ConnectionError:
        pytest.xfail('bad internet connection')

    assert (R/'thg_l1_asf_gako_2006093004_v01.cdf').is_file()
    assert (R/'thg_l2_asc_gako_19700101_v01.cdf').is_file()

    assert (R/'thg_l1_asf_fykn_2006093004_v01.cdf').is_file()
    assert (R/'thg_l2_asc_fykn_19700101_v01.cdf').is_file()


@pytest.mark.parametrize('time, site', [('2006-09-29T14', 'badname'),
                                        ('1950-01-01T01', 'gako')])
def test_nonexisting(time, site):
    try:
        tw.download(time, site, R, urls)
    except requests.exceptions.ConnectionError:
        pytest.xfail('bad internet connection')


if __name__ == '__main__':
    pytest.main([__file__])



#!/usr/bin/env python
"""
Plot time series for pixel(s) chosen by az/el

PlotThemisPixels ~/data/themis/ fykn 2011-01-06T17:00:00 -az 65 70 -el 48 68

PlotThemisPixels tests/ gako 2011-01-06T17:00:00 -az 65 70 -el 48 68
"""
from argparse import ArgumentParser
import numpy as np
import themisasi as ta
import themisasi.fov as taf
import themisasi.plots as tap
from matplotlib.pyplot import show


def main():
    p = ArgumentParser(description=' reads THEMIS GBO ASI CDF files and plays high speed video')
    p.add_argument('path', help='ASI data path')
    p.add_argument('site', help='site 4 character code e.g. gako')
    p.add_argument('treq', help='time (or time start,stop) requested', nargs='+')

    g = p.add_mutually_exclusive_group()
    p.add_argument('-az', help='azimuth(s) to plot (degrees)', type=float, nargs='+')
    g.add_argument('-el', help='elevation(s) to plot (degrees)', type=float, nargs='+')
    g.add_argument('-lla', help='latitude, longitude, altitude [km] projection',
                   type=float, nargs=3)

    p.add_argument('-o', '--odir', help='write video to this directory')
    p.add_argument('-v', '--verbose', action='store_true')
    P = p.parse_args()

    imgs = ta.load(P.path, site=P.site, treq=P.treq)

    if P.verbose:
        tap.plotazel(imgs)
# %% select nearest neighbor
    ind = taf.getimgind(imgs, P.lla, P.az, P.el)

    az, el, plat, plon, palt_m = taf.projected_coord(imgs, ind, P.lla)

    print(f'Using az, el {az}, {el}')
    print(f'Using projected lat,lon, alt [km]  {plat} {plon}  {palt_m}')

    dat = np.empty((imgs.time.size, ind.shape[0]), dtype=imgs['imgs'].dtype)
    for i, j in enumerate(ind):
        dat[:, i] = imgs['imgs'][:, j[0], j[1]]
# %% plot
    ttxt = f'{imgs.filename}'
    tap.plottimeseries(dat, imgs.time, ttxt)

    show()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Multi-camera view
GAKO: http://themis.ssl.berkeley.edu/data/themis/thg/l1/asi/gako/2008/03/
+ PFRR
"""
import themisasi as ta
import themisasi.plots as tap
from pathlib import Path
from matplotlib.pyplot import show
try:
    import seaborn as sns
    sns.set_context('talk')
except ImportError:
    pass

# %% user parameters
datadir = Path('~/data/themis').expanduser()
site = 'gako'
asifn = ('thg_l1_asf_gako_2008032607_v01.cdf',
         'thg_l1_asf_gako_2008032608_v01.cdf')
treq = [('2008-03-26T07:34', '2008-03-26T07:38'),  # Slide 9
        ('2008-03-26T07:46', '2008-03-26T07:50'),  # Slide 15
        ('2008-03-26T08:16', '2008-03-26T08:20'),  # Slide 30
        ]
odir = None

asical = datadir / 'themis_skymap_gako_20070401.sav'
# %% load data, with calibration az/el
data0 = ta.load(datadir / asifn[0], treq=treq[0], calfn=asical)
data1 = ta.load(datadir / asifn[0], treq=treq[1], calfn=asical)
data2 = ta.load(datadir / asifn[1], treq=treq[2], calfn=asical)
# %% plot
tap.plotasi(data0)
tap.plotasi(data1)
tap.plotasi(data2)

tap.plotazel(data0)

show()



#!/usr/bin/env python
"""
Plot THEMIS ASI in RA, DEC projection

ThemisRadec ~/data/2012-11-03/ gako 2012-11-03T06:30 ~/data/themis
"""
from argparse import ArgumentParser
import themisasi.io as tio
import themisasi.plots as tap


def main():
    p = ArgumentParser(description=' reads THEMIS GBO ASI CDF files and plays high speed video')
    p.add_argument('path', help='Themis ASI image file directory')
    p.add_argument('site', help='4 character site name e.g. gako')
    p.add_argument('treq', help='time or time range to load', nargs='+')
    p.add_argument('-c', '--calpath', help='path to calibration files', nargs='?')
    p.add_argument('-o', '--odir', help='output directory to write plots')
    P = p.parse_args()

    imgs = tio.load(P.path, site=P.site, treq=P.treq, calfn=P.calpath)
# %% plot
    tap.asi_radec(imgs, P.odir)

    print(imgs)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import pytest
import subprocess
from pathlib import Path

R = Path(__file__).parent


def test_video():
    pytest.importorskip('matplotlib')
    pytest.importorskip('pymap3d')
    subprocess.check_call(['PlotThemis', str(R), 'gako',
                           '2011-01-06T17:00:03'])


def test_pixels_azel():
    pytest.importorskip('histutils')
    pytest.importorskip('matplotlib')
    subprocess.check_call(['PlotThemisPixels', str(R), 'gako',
                           '2011-01-06T17:00:00', '2011-01-06T17:00:06',
                           '-az', '65', '70', '-el', '48', '68'])


# def test_pixels_latlon():
#    pytest.importorskip('histutils')
#    pytest.importorskip('matplotlib')
#    subprocess.check_call(['PlotThemisPixels', str(R), 'gako',
#                           '2011-01-06T17:00:00', '2011-01-06T17:00:06',
#                           '-lla', '68', '-145', '100.'])


if __name__ == '__main__':
    pytest.main(['-xrsv', __file__])



#!/usr/bin/env python
"""
example:
DownloadThemis 2012-11-03T06:23 2012-11-03T06:24 ~/data -s gako
"""
import themisasi.web as tweb
from argparse import ArgumentParser

VIDEO_BASE = 'http://themis.ssl.berkeley.edu/data/themis/thg/l1/asi/'
CAL_BASE = 'http://themis.ssl.berkeley.edu/data/themis/thg/l2/asi/cal/'


def main():
    p = ArgumentParser()
    p.add_argument('startend', help='start/end times UTC e.g. 2012-11-03T06:23 2012-11-03T06:24', nargs='+')
    p.add_argument('odir', help='directory to write downloaded CDF to')
    p.add_argument('-s', '--site', help='fykn gako etc.', nargs='+')
    p.add_argument('-overwrite', help='overwrite existing files', action='store_true')
    p.add_argument('-vh', help='Stem of URL for Video data', default=VIDEO_BASE)
    p.add_argument('-ch', help='Stem of URL for calibration data', default=CAL_BASE)

    P = p.parse_args()

    urls = {'video_stem': P.vh, 'cal_stem': P.ch}

    tweb.download(P.startend, P.site, P.odir, P.overwrite, urls)


if __name__ == '__main__':
    main()



import xarray
from pathlib import Path
import pymap3d as pm
import logging
import numpy as np
from matplotlib.pyplot import figure, draw, pause
from .plots import pcolormesh_nan, overlayrowcol


def asi_projection(dat: xarray.Dataset,
                   projalt_m: float = None,
                   min_el: float = 10.,
                   ofn: Path = None):
    """
    plots ASI projected to altitude

    * dat: Xarray containing image stack and metadata (az, el, lat, lon)
    * projalt_m: projection altitude in meters
    * min_el: minimum elevation angle (degrees). Data near the horizon is poorly calibrated (large angular error).
    * ofn: filename to write of plot (optional)
    """
    if projalt_m is None:
        logging.error('projection altitude must be specified')
        return

    if dat['imgs'].shape[0] == 0:
        return

    if ofn:
        ofn = Path(ofn).expanduser()
        odir = ofn.parent

# %% censor pixels near the horizon with large calibration error do to poor skymap fits
    badpix = dat['el'] < min_el
    az = dat['az'].values
    el = dat['el'].values
    az[badpix] = np.nan
    el[badpix] = np.nan
# %% coordinate transformation, let us know if error occurs
    slant_range = projalt_m / np.sin(np.radians(el))

    lat, lon, alt = pm.aer2geodetic(az, el, slant_range,
                                    dat.lat.item(), dat.lon.item(), dat.alt_m.item())
# %% plots
    fg = figure()
    ax = fg.gca()

    hi = pcolormesh_nan(lon, lat, dat['imgs'][0], cmap='gray', axis=ax)  # priming

    ttxt = f'Themis ASI {dat.site}  projected to altitude {projalt_m/1e3} km\n'  # FOV vs. HST0,HST1: green,red '
    ht = ax.set_title(ttxt, color='g')
    ax.set_xlabel('longitude')
    ax.set_ylabel('latitude')
    ax.autoscale(True, tight=True)
    ax.grid(False)
# %% plot narrow FOV outline
    if 'imgs2' in dat:
        overlayrowcol(ax, dat.rows, dat.cols)
# %% play video
    try:
        for im in dat['imgs']:
            ts = im.time.values.astype(str)[:-6]
            hi.set_array(im.values.ravel())  # for pcolormesh
            ht.set_text(ttxt + ts)
            draw()
            pause(0.01)
            if ofn:
                fn = odir / (ofn.stem + ts + ofn.suffix)
                print('saving', fn, end='\r')
                fg.savefig(fn, bbox_inches='tight', facecolor='k')
    except KeyboardInterrupt:
        return


def asi_radec(dat: xarray.Dataset,
              min_el: float = 10.,
              ofn: Path = None):
    """
    plots ASI projected to altitude

    * min_el: minimum elevation angle (degrees). Data near the horizon is poorly calibrated (large angular error).
    * ofn: filename to write of plot (optional)
    """

    if dat['imgs'].shape[0] == 0:
        return

    if ofn:
        ofn = Path(ofn).expanduser()

    az = dat.az.values
    el = dat.el.values
    az.setflags(write=True)
    el.setflags(write=True)

    bad = el < min_el  # low to horizon, calibration is very bad
    az[bad] = np.nan
    el[bad] = np.nan

    ra, dec = pm.azel2radec(az, el, dat.lat, dat.lon, dat.time)

    fg = figure()
    ax = fg.gca()

    pcolormesh_nan(ra, dec, dat['imgs'].values[0], cmap='gray', axis=ax)  # priming

    ttxt = f'Themis ASI {dat.site}  \n'
    ax.set_title(ttxt, color='g')
    ax.set_xlabel('right ascension [deg]')
    ax.set_ylabel('declination [deg]')
    ax.autoscale(True, tight=True)
    ax.grid(False)



#!/usr/bin/env python
from pathlib import Path
import pytest
from pytest import approx
import themisasi as ta
from datetime import datetime, timedelta, date
#
R = Path(__file__).parent
datfn = R / 'thg_l1_asf_gako_2011010617_v01.cdf'

cal1fn = R / 'themis_skymap_gako_20110305-+_vXX.sav'
cal2fn = R / 'thg_l2_asc_gako_19700101_v01.cdf'

assert datfn.is_file()
assert cal1fn.is_file()
assert cal2fn.is_file()


def test_missing_file(tmp_path):
    badfn = tmp_path / 'notafile.cdf'
    with pytest.raises(FileNotFoundError):
        ta.load(badfn)


def test_filename():
    data = ta.load(datfn)

    assert data['imgs'].site == 'gako'
    assert data['imgs'].shape == (23, 256, 256) and data['imgs'].dtype == 'uint16'


def test_filename_calname():

    with pytest.raises(ValueError):
        ta.load(datfn, calfn=cal1fn)

    data = ta.load(datfn, calfn=cal2fn)
    assert 'az' in data and 'el' in data
    assert data['az'].shape == data['imgs'].shape[1:]


def test_load_filename():
    """load by filename"""
    dat = ta.load(datfn)
    times = dat.time.values.astype('datetime64[us]').astype(datetime)

    assert (times >= datetime(2011, 1, 6, 17)).all()
    assert (times <= datetime(2011, 1, 6, 18)).all()


@pytest.mark.parametrize('site, time', [('gako', '2011-01-06T17:00:00'),
                                        ('gako', datetime(2011, 1, 6, 17))])
def test_load_site_time(site, time):
    """ load by sitename + time"""
    dat = ta.load(R, site, time)
    assert dat['imgs'].shape[0] == 1
    t = dat.time.values.astype('datetime64[us]').astype(datetime)
    assert abs(t - datetime(2011, 1, 6, 17, 0, 0)) < timedelta(seconds=.5)


@pytest.mark.parametrize('site, treq', [('gako', ('2011-01-06T17:00:00', '2011-01-06T17:00:12')),
                                        ('gako', ('2011-01-06T16:59:59', '2011-01-06T17:00:12'))])
def test_load_site_timerange(site, treq):
    """ load by sitename + timerange """
    dat = ta.load(R, site, treq=treq)
    assert dat['imgs'].shape[0] == 4
    times = dat.time.values.astype('datetime64[us]').astype(datetime)
    assert (times >= datetime(2011, 1, 6, 17)).all()
    assert (times <= datetime(2011, 1, 6, 17, 0, 12)).all()


@pytest.mark.parametrize('path, val, err', [(datfn, 1, TypeError),
                                            (datfn, '2011-01-06T16:59:59', ValueError),
                                            (datfn, '2011-01-06T18:00:01', ValueError),
                                            (R, '2010-01-01', FileNotFoundError),
                                            (R, ('2010-01-01', '2010-01-01T01'), FileNotFoundError)
                                            ])
def test_bad_time(path, val, err):
    with pytest.raises(err):
        ta.load(path, 'gako', treq=val)


def test_good_time():
    dat = ta.load(datfn, treq='2011-01-06T17:00:12')
    assert dat['imgs'].shape[0] == 1
    time = dat.time.values.astype('datetime64[us]').astype(datetime)
    assert abs(time - datetime(2011, 1, 6, 17, 0, 12)) < timedelta(seconds=0.02)

    dat = ta.load(datfn, treq=('2011-01-06T17:00:00', '2011-01-06T17:00:12'))
    assert dat['imgs'].shape[0] == 4
    time = dat.time.values.astype('datetime64[us]').astype(datetime)


def test_autoload_cal():
    dat = ta.load(R, 'gako', '2011-01-06T17:00:00')
    assert 'el' in dat and 'az' in dat


def test_calread_idl():

    cal = ta.loadcal(cal1fn)

    assert cal['el'][29, 161] == approx(15.458)
    assert cal['az'][29, 161] == approx(1.6255488)
    assert cal.lon == approx(-145.16)


def test_calread_cdf():
    cal = ta.loadcal(cal2fn)

    assert cal['el'][29, 161] == approx(19.132568)
    assert cal['az'][29, 161] == approx(183.81241)
    assert cal.lon == approx(-145.16)


def test_calread_sitedate():

    cal = ta.loadcal(R, 'gako', '2011-01-06')
    assert cal.caltime.date() == date(2007, 2, 2)


if __name__ == '__main__':
    pytest.main(['-xrsv', __file__])



#!/usr/bin/env python
"""
Using data from ICX249AL datasheet
and Jackel 2014
"""
import pandas
import numpy as np
from matplotlib.pyplot import figure, show

camfn = 'data/icx249al_response.csv'
filtfn = 'data/ir_filter.csv'

resp = pandas.read_csv(camfn, index_col=0)
filt = pandas.read_csv(filtfn, index_col=0)

wl = np.arange(400, 1000, 10)

# %% plot
ax = figure().gca()

resp.plot(ax=ax)
filt.plot(ax=ax)

ax.grid(True)
ax.set_title('THEMIS GBO ASI spectral response')
ax.set_xlabel('wavelength [nm]')
ax.set_ylabel('normalized')

show()



from pathlib import Path
import xarray
import numpy as np
from datetime import datetime
from matplotlib.pyplot import figure, draw, pause
from matplotlib.colors import LogNorm


def jointazel(cam: xarray.Dataset,
              ofn: Path = None,
              ttxt: str = ''):

    fg, axs = plotazel(cam, ttxt)
# %% plot line from other camera to magnetic zenith
    overlayrowcol(axs[0], cam.rows, cam.cols, 'g')
    overlayrowcol(axs[1], cam.rows, cam.cols, 'g')
# %% plot plane including this camera and line to magnetic zenith from other camera
    try:
        axs[0] = overlayrowcol(axs[0], cam.cutrow, cam.cutcol, 'r', '1-D cut')
        axs[1] = overlayrowcol(axs[1], cam.cutrow, cam.cutcol, 'r', '1-D cut')
    except AttributeError:
        pass  # this was not a 1-D plane case

    if ofn:
        ofn = Path(ofn).expanduser()
        print('saving', ofn)
        fg.savefig(ofn, bbox_inches='tight', dpi=100)

    if 'Baz' in cam.attrs:
        axs[0].scatter(cam.Bcol, cam.Brow, 250, 'red', '*', label='Mag. Zen.')
        axs[1].scatter(cam.Bcol, cam.Brow, 250, 'red', '*', label='Mag. Zen.')

    for a in axs:
        a.legend()


def plotazel(data: xarray.Dataset, ttxt: str = ''):

    if 'az' not in data or 'el' not in data:
        return

    fg = figure(figsize=(12, 6))
    ax = fg.subplots(1, 2, sharey=True)

    ttxt = f'{data.filename} {ttxt}'
    if data.caltime is not None:
        ttxt += f'{data.caltime}'

    fg.suptitle(ttxt)

    c = ax[0].contour(data['az'].x, data['az'].y, data['az'])
    ax[0].clabel(c, fmt='%0.1f')
    ax[0].set_title('azimuth')
    ax[0].set_xlabel('x-pixels')
    ax[0].set_ylabel('y-pixels')

    c = ax[1].contour(data['el'].x, data['el'].y, data['el'])
    ax[1].clabel(c, fmt='%0.1f')
    ax[1].set_title('elevation')
    ax[1].set_xlabel('x-pixels')

    return fg, ax


def plottimeseries(data: np.ndarray, time: datetime, ttxt: str = ''):

    assert data.ndim == 2

    ax = figure().gca()

    ax.plot(time, data)
    ax.set_xlabel('time')
    ax.set_ylabel('brightness [data numbers]')
    ax.set_title(ttxt)


def overlayrowcol(ax, rows, cols, color: str = None, label: str = None):
    """
    plot FOV outline onto image via the existing axis "ax"

    inputs:
    ax: existing plot axis to overlay lines outlining FOV
    rows,cols: indices to plot
    """
    if rows is None or cols is None:
        return

    if len(rows) == 1 or isinstance(rows, (np.ndarray, xarray.DataArray)) and rows.ndim == 1:
        ax.scatter(cols, rows, color=color, alpha=0.5, marker='.', label=label)
    else:
        raise ValueError('unknonn row/col layout, was expecting 1-D')

    return ax
# %%


def plotasi(data: xarray.Dataset, ofn: Path = None):
    """
    rows,cols expect lines to be along rows Nlines x len(line)
    list of 1-D arrays or 2-D array
    """

    if data['imgs'].shape[0] == 0:
        return

    if ofn:
        ofn = Path(ofn).expanduser()
        odir = ofn.parent

    fg = figure()
    ax = fg.gca()

    hi = ax.imshow(data['imgs'][0], cmap='gray', origin='lower',
                   norm=LogNorm(), interpolation='none')  # priming
    ttxt = f'Themis ASI {data.site}\n'  # FOV vs. HST0,HST1: green,red '
    ht = ax.set_title(ttxt, color='g')
    ax.set_xlabel('x-pixels')
    ax.set_ylabel('y-pixels')
    ax.autoscale(True, tight=True)
    ax.grid(False)
# %% plot narrow FOV outline
    if 'imgs2' in data:
        overlayrowcol(ax, data.rows, data.cols)
# %% play video
    try:
        for im in data['imgs']:
            ts = str(im.time.astype('datetime64[us]').astype(datetime))
            hi.set_data(im)
            ht.set_text(ttxt + ts)
            draw()
            pause(0.01)
            if ofn:
                fn = odir / (ofn.stem + ts + ofn.suffix)
                print('saving', fn, end='\r')
                fg.savefig(fn, bbox_inches='tight', facecolor='k')
    except KeyboardInterrupt:
        return


def pcolormesh_nan(x: np.ndarray, y: np.ndarray, c: np.ndarray, cmap=None, axis=None):
    """handles NaN in x and y by smearing last valid value in column or row out,
    which doesn't affect plot because "c" will be masked too
    """

    mask = np.isfinite(x) & np.isfinite(y)
    top = None

    for i, m in enumerate(mask):
        good = m.nonzero()[0]

        if good.size == 0:
            continue
        elif top is None:
            top = i
        else:
            bottom = i

        x[i, good[-1]:] = x[i, good[-1]]
        y[i, good[-1]:] = y[i, good[-1]]

        x[i, :good[0]] = x[i, good[0]]
        y[i, :good[0]] = y[i, good[0]]

    x[:top, :] = np.nanmean(x[top, :])
    y[:top, :] = np.nanmean(y[top, :])

    x[bottom:, :] = np.nanmean(x[bottom, :])
    y[bottom:, :] = np.nanmean(y[bottom, :])

    if axis is None:
        axis = figure().gca()

    return axis.pcolormesh(x, y, np.ma.masked_where(~mask, c), cmap=cmap)
